# Copyright The Lightning AI team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import multiprocessing as mp

from litserve.loops.base import LitLoop, _BaseLoop
from litserve.loops.continuous_batching_loop import ContinuousBatchingLoop, Output
from litserve.loops.loops import (
    get_default_loop,
    inference_worker,
)
from litserve.loops.simple_loops import BatchedLoop, SingleLoop
from litserve.loops.streaming_loops import BatchedStreamingLoop, StreamingLoop

mp.allow_connection_pickling()

__all__ = [
    "_BaseLoop",
    "LitLoop",
    "get_default_loop",
    "inference_worker",
    "Output",
    "SingleLoop",
    "BatchedLoop",
    "StreamingLoop",
    "BatchedStreamingLoop",
    "ContinuousBatchingLoop",
]
