{%- set external_urls = {
  'github': 'https://github.com/Lightning-AI/utilities',
  'github_issues': 'https://github.com/Lightning-AI/utilities/issues',
  'contributing': 'https://github.com/Lightning-AI/lightning/blob/master/CONTRIBUTING.md',
  'governance': 'https://github.com/Lightning-AI/lightning/blob/master/governance.md',
  'docs': 'https://dev-toolbox.rtfd.io/en/latest',
  'twitter': 'https://twitter.com/PyTorchLightnin',
  'discuss': 'https://pytorch-lightning.slack.com',
  'home': 'https://lightning-tools.rtfd.io/en/latest/',
  'get_started': 'https://lightning-tools.readthedocs.io/en/latest/introduction_guide.html',
  'features': 'https://lightning-tools.rtfd.io/en/latest/',
  'blog': 'https://www.pytorchlightning.ai/blog',
  'support': 'https://lightning-tools.rtfd.io/en/latest/',
}
-%}
